#!/usr/bin/env python
import argparse
import omicidx.sra_parsers
import json
import logging
import collections
from xml.etree import ElementTree as et


def main():
    logging.basicConfig(
        level=logging.INFO, format="%(asctime)s %(name)-12s %(levelname)-8s %(message)s"
    )
    logger = logging.getLogger(__name__)

    parser = argparse.ArgumentParser(
        description="""
    Transforms an SRA XML mirroring metadata file into
    corresponding JSON format files. JSON is line-delimited
    JSON (not an array)."""
    )

    parser.add_argument("fname", help="The SRA xml filename")
    parser.add_argument(
        "outfile", help="the output filename where to write json records"
    )

    opts = parser.parse_args()

    entity = ""

    if "study" in opts.fname:
        entity = "STUDY"
        sra_parser = omicidx.sra_parsers.SRAStudyRecord

    if "run" in opts.fname:
        entity = "RUN"
        sra_parser = omicidx.sra_parsers.SRARunRecord

    if "sample" in opts.fname:
        entity = "SAMPLE"
        sra_parser = omicidx.sra_parsers.SRASampleRecord

    if "experiment" in opts.fname:
        entity = "EXPERIMENT"
        sra_parser = omicidx.sra_parsers.SRAExperimentRecord

    logger.info("configuration: {}".format(str(opts)))
    logger.info("using {} entity type".format(entity))
    logger.info("parsing {} records".format(entity))
    n = 0
    with open(opts.outfile, "w") as outfile:
        with omicidx.sra_parsers.open_file(opts.fname) as f:
            for event, element in et.iterparse(f):
                if event == "end" and element.tag == entity:
                    rec = sra_parser(element).data
                    n += 1
                    if (n % 100000) == 0:
                        logger.info("parsed {} {} entries".format(entity, n))
                    outfile.write(json.dumps(rec) + "\n")
                    element.clear()
            logger.info("parsed {} entity entries".format(n))


if __name__ == "__main__":
    main()



"""Usage

Common use will be to get a list of accessions using:



"""

from Bio import Entrez
import logging
import collections
import re
import datetime
import urllib
from xml import etree
import xml
from io import StringIO
import requests
import httpx
from tenacity import retry
from tenacity.wait import wait_exponential
from tenacity.stop import stop_after_attempt
from urllib.error import HTTPError  # for Python 3

import omicidx.geo.pydantic_models as pydantic_models

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("omicidx")

ETYP = ["GSE", "GSM", "GPL", "GDS"]


def get_entrez_instance(email="user@example.com"):
    """Return a Bio::Entrez object

    # Arguments
    email (str): the email to be used with the Entrez instance

    Returns
    =======
    A Bio::Entrez instance
    """
    ret = Entrez
    ret.email = email
    return ret


def get_geo_accessions(
    etyp=None, batch_size=1000, add_term=None, email="user@example.com"
):
    """Get GEO accessions by etyp

    Useful for getting all the ETYP accessions for
    later bulk processing

    Parameters
    ----------
    etyp: str
        One of GSE, GPL, GSM, GDS
    batch_size: int
        the number of accessions to return in one batch.
        Transparent to the user, as this returns an iterator.
    add_term: str
        Add a search term for the query. Useful to limit
        by date or search for specific text. For example,
        to limit by date: '2007/01/01:2007/03/01[PDAT]'
    email: str
        user email (not important)

    Return
    ------
    an iterator of accessions, each as a string
    """
    term = ""
    if add_term is not None:
        term = term + add_term
    entrez = get_entrez_instance(email)
    if etyp is not None:
        if etyp not in ETYP:
            raise Exception(
                "etype {} not one of the accepted etyps {}".format(etyp, str(ETYP))
            )
        term = term + " " + etyp + "[ETYP]"
    if term != "":

        handle = entrez.esearch(db="gds", term=term, usehistory="y")
    else:
        raise Exception("term is required for entrez search access")
    search_results = entrez.read(handle)
    webenv = search_results["WebEnv"]
    query_key = search_results["QueryKey"]
    count = int(search_results["Count"])
    data = []
    logging.info("found {} records for {} database".format(count, etyp))
    n = 0
    for start in range(0, count, batch_size):
        end = min(count, start + batch_size)
        attempt = 0
        fetch_handle = None
        while attempt < 10:
            attempt += 1
            try:
                fetch_handle = entrez.esummary(
                    db="gds",
                    rettype="acc",
                    retmode="xml",
                    retstart=start,
                    retmax=batch_size,
                    webenv=webenv,
                    query_key=query_key,
                )
                for g in entrez.read(fetch_handle):
                    n += 1
                    yield (g["Accession"])

                break
            except (HTTPError, RuntimeError) as err:
                import time

                if isinstance(err, HTTPError):
                    logging.error("Received error from server %s" % err)
                    logging.error("Attempt %i of 10" % attempt)
                    time.sleep(1 * attempt * attempt)
                else:
                    logging.error("RuntimeError received: {}".format(err))
                    logging.error("Attempt %i of 10" % attempt)
                    time.sleep(1)
            else:
                raise


def get_geo_accession_xml(accession, targ="all", view="brief"):
    """Open a connection to get the GEO XML for an accession

    Parameters
    ==========
    accession: str the GEO accesssion
    targ: str what to fetch. One of "all", "series", "platform",
        "samples", "self"
    view: str amount to return.

    Returns
    =======
    A file-like object for reading or readlines

    >>> handle = get_geo_accession_xml('GSE2553')
    >>> handle.readlines()
    """
    url = "https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?targ={}&acc={}&form=xml&view={}".format(
        targ, accession, view
    )
    attempt = 0
    while attempt < 10:
        attempt += 1
        try:
            return urllib.request.urlopen(url)
            break
        except Exception as err:
            print("Received error from server %s" % err)
            print("Attempt %i of 3" % attempt)
            import time

            time.sleep(1 * attempt)


def get_geo_accession_soft(
    accession: str, targ: str = "all", view: str = "brief"
) -> str:
    """Open a connection to get the GEO SOFT for an accession

    Parameters
    ==========
    accession: str the GEO accesssion
    targ: str what to fetch. One of "all", "series", "platform",
        "samples", "self"
    view: str amount to return.

    Returns
    =======
    A file-like object for reading or readlines

    >>> handle = get_geo_accession_soft('GSE2553')
    >>> handle.readlines()
    """
    logging.info(f"Accessing accession {accession} in SOFT format")
    url = "https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?targ={}&acc={}&form=text&view={}".format(
        targ, accession, view
    )

    # @retry(
    #     wait=wait_exponential(multiplier=1, min=10, max=60), stop=stop_after_attempt(30)
    # )
    def _perform_get(url: str) -> str:
        resp = httpx.get(url, timeout=120)
        resp.raise_for_status()
        return resp.text

    res = _perform_get(url)
    return res


def _split_on_first_equal(line, val="="):
    """Split a line based on first occurrence of " ="

    Parameters
    ----------
    line: str
        string to split
    split: str
        split based on this. Can be multiple characters.
    >>> _split_on_first_equal("this = abc = 1")
    ('this', 'abc = 1')
    """
    l = line.split(" {} ".format(val))
    if l[0].endswith(" {}".format(val)):
        l[0] = l[0].replace(" {}".format(val), "")
    return (l[0], " {} ".format(val).join(l[1:]))


def get_geo_entities(txt):
    """Get the text associated with each entity from a block of text

    Parameters
    ----------
    txt: list(str)

    Returns
    -------
    A dict of entities keyed by accession and values a list of txt lines
    """
    logging.info("Parsing out geo entities")
    entities = {}
    accession = None
    entity = []
    for line in txt:
        line = line.strip()
        if line.startswith("^"):
            accession = _split_on_first_equal(line)[1]
            entities[accession] = entity
            entity = []
        entity.append(line)
    entities[accession] = entity
    logging.info(f"found {len(entities.keys())}")
    return entities


#####################################
# Utility functions for GSE parsing #
#####################################


def get_subseries_from_relations(relation_list):
    ret = []
    for i in relation_list:
        if i.startswith("SuperSeries of:"):
            ret.append(i.replace("SuperSeries of: ", ""))
    return ret


def get_bioprojects_from_relations(relation_list):
    ret = []
    for i in relation_list:
        if i.startswith("BioProject: https://www.ncbi.nlm.nih.gov/bioproject/"):
            ret.append(
                i.replace("BioProject: https://www.ncbi.nlm.nih.gov/bioproject/", "")
            )
    return ret


def get_SRA_from_relations(relation_list):
    ret = []
    for i in relation_list:
        if i.startswith("SRA: https://www.ncbi.nlm.nih.gov/sra?term="):
            ret.append(i.replace("SRA: https://www.ncbi.nlm.nih.gov/sra?term=", ""))
    return ret


def get_biosample_from_relations(relation_list):
    ret = []
    for i in relation_list:
        if i.startswith("BioSample: https://www.ncbi.nlm.nih.gov/biosample/"):
            ret.append(
                i.replace("BioSample: https://www.ncbi.nlm.nih.gov/biosample/", "")
            )
    return ret


def get_channel_characteristics(d, ch):
    ch_items = list([k for k in d.keys() if k.endswith("ch{}".format(ch))])
    characteristics = []
    ret = {}
    for k in ch_items:
        if k.startswith("characteristics"):
            for characteristic in d[k]:
                characteristics.append(tuple(characteristic.split(": ")))
            continue
        newk = k.replace("_ch{}".format(ch), "")
        ret[k.replace("_ch{}".format(ch), "")] = "\n".join(d[k])
        if newk == "taxid":
            ret[newk] = list([int(x) for x in ret[newk].split("\n")])
    char = []
    for v in characteristics:
        if len(v) == 1:
            tag = v[0]
            val = None
        elif len(v) == 2:
            tag = v[0]
            val = v[1]
        elif len(v) > 2:
            tag = v[0]
            val = ":".join(v[1:])
        char.append({"tag": tag, "value": val})
    ret["characteristics"] = char
    return ret


def _split_geo_name(v):
    """split name of form first,middle,last into dict"""
    return dict(zip("first middle last".split(), v.split(",")))


def _create_contact_from_parsed(d):
    contact_dict = {}
    for k, v in d.items():
        if k.startswith("contact"):
            v = v[0]
            if v == "":
                v = None
            if k == "contact_name":
                v = _split_geo_name(v)
            if k == "contact_web_link":
                if v is not None and type(v) == str and len(v):
                    if not str(v).startswith("http"):
                        v = "http://" + str(v)
            contact_dict[k.replace("contact_", "")] = v
    return contact_dict


def _split_contributor_names(contribs):
    if len(contribs) == 0:
        return []
    return list([_split_geo_name(v) for v in contribs])


###############################################
# Parse a single entity of SOFT format,       #
# generically. Then pass along to             #
# GSE, GSM, or GPL specific parser and return #
# appropriate class.                          #
###############################################
def _parse_single_entity_soft(entity_txt):
    # Deal with common stuff first:
    tups = [_split_on_first_equal(line) for line in entity_txt]
    accession = tups[0][1]

    entity_type = tups[0][0].replace("^", "")
    tups = [(re.sub("![^_]+_", "", tup[0]), tup[1]) for tup in tups]
    d = collections.defaultdict(list)
    for tup in tups[1:]:
        d[tup[0]].append(tup[1])
    d2 = dict((k, v) for k, v in d.items())
    # int fields
    INT_FIELDS = ["pubmed_id", "sample_taxid", "platform_taxid"]
    for i in INT_FIELDS:
        if i in d2:
            try:
                d2[i] = list(map(lambda x: int(x), d2[i]))
            except:
                d2[i] = []
    # change "geo_accesion = [...]" to "accession = ..."
    if "geo_accession" in d2:
        d2["accession"] = d2["geo_accession"][0]
    elif "accession" in d2:
        d2["accession"] = d2["accession"][0]
    else:
        d2["accession"] = "UNKNOWN"  # type: ignore
    if "description =" in d2:
        del d2["description ="]
    for k in [
        "status",
        "description",
        "data_processing",
        "title",
        "summary",  # GSE only?
        "submission_date",  # to fix later
        "last_update_date",  # to fix later
        "overall_design",
    ]:  # GSE only ]:
        try:
            d2[k] = "\n".join(d2[k])
        except KeyError:
            d2[k] = None
    # contact details as object
    contact = _create_contact_from_parsed(d2)
    dkeys = d2.keys()
    for k in list(dkeys):
        if k.startswith("contact"):
            del d2[k]
    d2["contact"] = contact
    if entity_type == "SERIES":
        return _parse_single_gse_soft(d2)
    if entity_type == "SAMPLE":
        return _parse_single_gsm_soft(d2)
    if entity_type == "PLATFORM":
        return _parse_single_gpl_soft(d2)


###############################################################################
#        Update date fields of format 'month day year' to date type.          #
###############################################################################
def _fix_date_fields(d):
    for datefield in ["submission_date", "last_update_date"]:
        if datefield in d:
            d[datefield] = datetime.datetime.strptime(d[datefield], "%b %d %Y")
    return d


#######################################
# Parse the GSE entity in SOFT format #
#######################################
def _parse_single_gse_soft(d2):
    try:
        d2["subseries"] = get_subseries_from_relations(d2["relation"])
        d2["bioprojects"] = get_bioprojects_from_relations(d2["relation"])
        d2["sra_studies"] = get_SRA_from_relations(d2["relation"])
        # GEO sometimes references SRS and SRR from series--filter those out
        d2["sra_studies"] = list(filter(lambda a: a.find("P") > 0, d2["sra_studies"]))
    except KeyError:
        d2["subseries"] = []
        d2["bioprojects"] = []
        d2["sra_studies"] = []
    supp_files = []
    for k in list(d2.keys()):
        if k.startswith("supplementa"):
            supp_files += d2[k]
            del d2[k]
    d2["supplemental_files"] = supp_files
    if len(supp_files) > 0:
        if supp_files[0] == "NONE":
            supp_files = []
    d2["_entity"] = "GSE"
    if "contributor" in d2:
        d2["contributor"] = _split_contributor_names(d2["contributor"])
    d2 = _fix_date_fields(d2)
    return pydantic_models.GEOSeries(**d2)


def _create_gsm_channel_data(d):
    ch_count = int(d["channel_count"])
    ch_recs = []
    for i in range(0, ch_count):
        # this GEOChannel just encapsulates the parsing
        # of channel records.
        # TODO: refactor to parse directly to dict
        ch_recs.append(get_channel_characteristics(d, i + 1))
    return ch_recs


# Search for fields like _ch1 and _ch2 ....
import re

r1 = re.compile(r"_ch\d+$")


#######################################
# Parse the GSM entity in SOFT format #
#######################################
def _parse_single_gsm_soft(d2):
    # to singular:
    for k in [
        "platform_id",
        "channel_count",
        "library_strategy",
        "library_source",
        "library_selection",
        "instrument_model",
        "data_row_count",
        "anchor",
        "tag_length",
        "type",
        "tag_count",
    ]:
        try:
            d2[k] = "\n".join(d2[k])
        except KeyError:
            d2[k] = None
    try:
        d2["biosample"] = get_biosample_from_relations(d2["relation"])[0]
        d2["sra_experiment"] = get_SRA_from_relations(d2["relation"])[0]
    except:
        d2["biosample"] = None
        d2["sra_experiment"] = None
    if "contributor" in d2:
        d2["contributor"] = _split_contributor_names(d2["contributor"])
    d2["channels"] = _create_gsm_channel_data(d2)
    supp_files = []
    for k in list(d2.keys()):
        if r1.search(k):  # have to use search here because endswith does not take regex
            del d2[k]
        if k.startswith("supplementa"):
            supp_files += d2[k]
            del d2[k]
    d2["supplemental_files"] = supp_files
    if supp_files[0] == "NONE":
        supp_files = []
    d2["_entity"] = "GSM"
    d2 = _fix_date_fields(d2)
    # Collapse protocol fields into single string
    for k in filter(lambda k: k.endswith("protocol"), d2.keys()):
        if isinstance(d2[k], list):
            d2[k] = " ".join(d2[k])
    if "contributor" in d2:
        d2["contributor"] = _split_contributor_names(d2["contributor"])
    return pydantic_models.GEOSample(**d2)


#######################################
# Parse the GPL entity in SOFT format #
#######################################
def _parse_single_gpl_soft(d2):
    for k in ["technology", "distribution", "organism", "taxid", "data_row_count"]:
        try:
            d2[k] = "\n".join(d2[k])
        except KeyError:
            d2[k] = None
    d2["_entity"] = "GPL"
    d2 = _fix_date_fields(d2)
    if "manufacture_protocol" in d2:
        d2["manufacture_protocol"] = "\n".join(d2["manufacture_protocol"])
    if "contributor" in d2:
        d2["contributor"] = _split_contributor_names(d2["contributor"])
    return pydantic_models.GEOPlatform(**d2)


###############################################################################
#                   This is the main entrypoint for parsing.                  #
###############################################################################
def geo_entity_iterator(geo: str, targ: str = "self", view: str = "brief"):
    """Returns an iterator of GEO entities

    Given a GEO accession (typically a GSE,
    will return an iterator of the GEO entities
    associated with the record, including all
    GSMs, GPLs, and the GSE record itself

    Parameters
    ----------
    fh: Iterable[str]
       Anything that can iterate over lines of text
       Could be a list of text lines, a file handle, or
       an open url.

    Yields
    ------
    Iterator of GEO entities as pydantic models


    >>> for i in geo_soft_entity_iterator('GSE2553'):
    ...     print(i)
    """
    entity = []
    accession = None
    in_table = False
    logger.info(f"Getting GEO entities for {geo}")
    text = get_geo_accession_soft(geo, targ=targ, view=view)
    logger.info(f"Got GEO entities for {geo}")
    for line in StringIO(text):
        try:
            if isinstance(line, bytes):
                line = line.decode()
        except:
            pass
        line = line.strip()

        # ignore table details for now

        if line.endswith("table_begin"):
            in_table = True
        if line.endswith("table_end"):
            in_table = False
            continue
        if in_table:
            continue

        # ignore comments

        if line.startswith("#"):
            continue

        if line.startswith("^"):
            if accession is not None:
                logger.info(f".. parsing found {accession}")
                yield (_parse_single_entity_soft(entity))
            accession = _split_on_first_equal(line)[1]
            entity = []

        entity.append(line)
    yield (_parse_single_entity_soft(entity))


def gse_to_json(gse):
    """Return json version of GSE record as text

    Parameters
    ----------
    gse : str
        The GSE accession to convert
    """
    res = []
    for z in geo_entity_iterator(gse):
        res.append(z.json())
    return "\n".join(res)


def bulk_gse_to_json(term, outfile="/dev/stdout"):
    """Bulk process GSEs to json

    This function takes a term (see below) and does a search. Results
    from the search of GSE records are returned as json lines to the
    output file in `outfile`.

    Parameters
    ----------
    term : str
        A search string like "2019/01/28:2019/02/28[PDAT]" as an example.
        Passed to entrez search.
    outfile: str
        The output file to which to write
    """
    for gse in get_geo_accessions(add_term=term):
        if outfile is not None:
            logging.info(f"writing accession {gse}")
            res = gse_to_json(gse)
            if res is not None:
                outfile.write(res + "\n")
            else:
                logging.warning(f"Accession {gse} not found, it appears.")


def gse_accessions(term, outfile="/dev/stdout"):
    """Output list of gse accessions based on term search

    Writes to a file.

    Parameters
    ----------
    term : str
        A search string like "2019/01/28:2019/02/28[PDAT]" as an example.
        Passed to entrez search.
    outfile: str
        The output file to which to write

    """

    with open(outfile, "w") as ofile:
        n = 0
        for gse in get_geo_accessions(add_term=term, batch_size=1000):
            n += 1
            if n % 5000 == 0:
                logging.info(f"Processed {n} records")
            ofile.write(gse + "\n")
        logging.info(f"Processed {n} records")



"""SRA parsers including:

- study
- sample
- experiment
- run

These parsers each parse XML format files of the format
available from the fullxml api.

The main entry point into this module is the `parse_xml_file`
function.

"""

import re
from collections import defaultdict
import urllib
import urllib.error
import urllib.request
import xml.etree.ElementTree as etree
import gzip
from . import pydantic_models
import logging

logger = logging.getLogger("sra_parser")


def parse_xml_url(url: str, entity: str, gz: bool = True):
    n = 0

    with gzip.open(urllib.request.urlopen(url), "rb") as f:
        for event, element in etree.iterparse(f):
            if event == "end" and element.tag == entity:
                rec = globals()["SRA" + entity.title() + "Record"](element).data
                n += 1
                if (n % 100000) == 0:
                    logger.info("parsed {} {} entries".format(entity, n))
                element.clear()
                yield (rec)
    logger.info("parsed {} entity entries".format(n))


def parse_xml_file(xmlfilename):
    """Parse an NCBI SRA mirroring XML file

    This function returns an iterator over the
    records in the xml file, returning a dict
    of parsed records.

    For example:

    wget --mirror -nH --cut-dirs=3 ftp://ftp.ncbi.nlm.nih.gov/sra/reports/Mirroring/NCBI_SRA_Mirroring_20181027/

    >>> import omicidx.sra_parsers as sp
    >>> studies = sp.parse_xml_file("NCBI_SRA_Mirroring_20181027/meta_study_set.xml.gz")
    >>> next(studies)
    ...


    Parameters
    ----------
    xmlfilename : string
        the filename to be parsed. Can be gzipped. Must include
        the "entity" name in the filename (eg., "run", "experiment")

    Returns
    -------
    iterator:
        An iterator of dict records from parsing each xml record.

    """
    if "study" in xmlfilename:
        entity = "STUDY"
        sra_parser = SRAStudyRecord
    if "run" in xmlfilename:
        entity = "RUN"
        sra_parser = SRARunRecord
    if "sample" in xmlfilename:
        entity = "SAMPLE"
        sra_parser = SRASampleRecord
    if "experiment" in xmlfilename:
        entity = "EXPERIMENT"
        sra_parser = SRAExperimentRecord
    n = 0
    with open_file(xmlfilename) as f:
        for event, element in etree.iterparse(f):
            if event == "end" and element.tag == entity:
                rec = sra_parser(element).data
                n += 1
                if (n % 100000) == 0:
                    logger.info("parsed {} {} entries".format(entity, n))
                element.clear()
                yield (rec)
    logger.info("parsed {} entity entries".format(n))


def parse_study(xml):
    """Parse an SRA xml STUDY element

    Parameters
    ----------
    xml: an xml.etree Element

    Returns
    -------
    A dict object parsed from the XML
    """

    required_keys = [
        "abstract",
        "BioProject",
        "GEO",
        "accession",
        "alias",
        "attributes",
        "center_name",
        "broker_name",
        "description",
        "study_type",
        "study_accession",
        "title",
    ]
    d = dict((k, None) for k in required_keys)
    try:
        d.update(xml.attrib)
    except AttributeError:
        pass
    path_map = {
        "title": (".//STUDY_TITLE", "text"),
        "abstract": (".//STUDY_ABSTRACT", "text"),
        "description": (".//STUDY_DESCRIPTION", "text"),
    }
    d.update(_process_path_map(xml, path_map))
    d.update(_parse_identifiers(xml.find("IDENTIFIERS"), "study"))
    try_update(d, _parse_study_type(xml.find("DESCRIPTOR/STUDY_TYPE")))
    d = try_update(d, _parse_attributes(xml.find("STUDY_ATTRIBUTES")))
    d.update(_parse_links(xml.find("STUDY_LINKS")))
    pubmeds = []
    if "xrefs" in d:
        for xref in d["xrefs"]:
            if xref["db"] == "pubmed":
                if xref["id"] is not None:
                    pubmeds.append(xref["id"])
    d.update({"pubmed_ids": pubmeds})
    return d


def parse_submission(xml):
    """Parse an SRA xml SUBMISSION element

    Parameters
    ----------
    xml: xml.etree.ElementTree.Element

    Returns
    -------
    a dict of experiment
    """
    d = {}
    d.update(xml.attrib)
    d.update(_parse_identifiers(xml.find("IDENTIFIERS"), "submission"))
    return d


def dict_from_single_xml(txt):
    xml = etree.fromstring(txt)
    entity = xml.tag.lower()
    vals = globals()["parse_" + entity](xml)
    vals["entity_type"] = xml.tag.lower()
    return vals


def model_from_single_xml(txt):
    xml = etree.fromstring(txt)
    entity = xml.tag.lower()
    et = globals()["parse_" + entity](xml)
    return getattr(pydantic_models, "Sra" + entity.capitalize())(**et)


def parse_run(xml):
    """Parse an SRA xml RUN element

    Parameters
    ----------
    xml: an xml.etree Element

    Returns
    -------
    A dict object parsed from the XML
    """
    d = {}
    d.update(xml.attrib)
    for k in ["total_spots", "total_bases", "size"]:
        try:
            d[k] = int(d[k])
        except:
            # missing some key elements
            pass
    try:
        d["avg_length"] = float(d["total_bases"]) / d["total_spots"]
    except:
        pass
    path_map = {
        "experiment_accession": ("EXPERIMENT_REF", "accession"),
        "title": ("TITLE", "text"),
    }

    d = try_update(d, _parse_taxon(xml.find("tax_analysis")))
    # d = try_update(d, _parse_run_reads(xml.find(".//SPOT_DESCRIPTOR")))
    d.update(_process_path_map(xml, path_map))
    d.update(_parse_identifiers(xml.find("IDENTIFIERS"), "run"))
    d = try_update(d, _parse_attributes(xml.find("RUN_ATTRIBUTES")))
    d = try_update(d, _parse_attributes(xml.find("RUN_ATTRIBUTES")))
    d = try_update(d, _parse_run_files(xml.find("SRAFiles")))
    d = try_update(d, _parse_run_stats(xml.find("Statistics")))
    d = try_update(d, _parse_run_bases(xml.find("Bases")))
    d = try_update(d, _parse_run_qualities(xml))
    try:
        # already have accession, so no need for this
        del d["run_accession"]
    except:
        pass
    return d


def _parse_run_stats(xml):
    if xml is None:
        return None
    stats = []
    for read in xml.findall("Read"):
        ret = {}
        ret["index"] = int(read.get("index", 0))
        ret["count"] = int(read.get("count", 0))
        ret["mean_length"] = float(read.get("average", 0.0))
        ret["sd_length"] = float(read.get("stdev", 0.0))
        stats.append(ret)
    return {"reads": stats}


def _parse_run_bases(xml):
    if xml is None:
        return None
    ret = []
    for base in xml.findall("Base"):
        ret.append({base.get("value"): int(base.get("count"))})
    return {"base_counts": ret}


def _parse_run_files(xml):
    if xml is None:
        return None
    files = xml.findall("./SRAFile")
    ret = []
    for f in files:
        retfile = {}
        for k in f.keys():
            retfile[k] = f.get(k)
        retfile["alternatives"] = []
        for alt in f.findall("Alternatives"):
            altfile = {}
            for k in alt.keys():
                altfile[k] = alt.get(k)
            retfile["alternatives"].append(altfile)
        ret.append(retfile)
    return {"files": ret}


def parse_experiment(xml):
    """Parse an SRA xml EXPERIMENT element

    Parameters
    ----------
    xml: xml.etree.ElementTree.Element

    Returns
    -------
    a dict of experiment
    """
    required_keys = [
        "accession",
        "attributes",
        "alias",
        "center_name",
        "design",
        "description",
        "experiment_accession",
        "identifiers",
        "instrument_model",
        "library_name",
        "library_construction_protocol",
        "library_layout_orientation",
        "library_layout_length",
        "library_layout_sdev",
        "library_strategy",
        "library_source",
        "library_selection",
        "library_layout",
        "xrefs",
        "platform",
        "sample_accession",
        "study_accession",
        "title",
    ]

    d = dict((k, None) for k in required_keys)
    try:
        d.update(xml.attrib)
    except:
        import xml.etree.ElementTree as et

        et.tostring(xml)

    path_map = {
        "title": ("./TITLE", "text"),
        "study_accession": ("./STUDY_REF/IDENTIFIERS/PRIMARY_ID", "text"),
        "design": ("./DESIGN/DESIGN_DESCRIPTION", "text"),
        "library_name": ("./DESIGN/LIBRARY_DESCRIPTOR/LIBRARY_NAME", "text"),
        "library_strategy": ("./DESIGN/LIBRARY_DESCRIPTOR/LIBRARY_STRATEGY", "text"),
        "library_source": ("./DESIGN/LIBRARY_DESCRIPTOR/LIBRARY_SOURCE", "text"),
        "library_selection": ("./DESIGN/LIBRARY_DESCRIPTOR/LIBRARY_SELECTION", "text"),
        "library_layout": (
            "./DESIGN/LIBRARY_DESCRIPTOR/LIBRARY_LAYOUT",
            "child",
            "tag",
        ),
        "library_layout_orientation": (
            "./DESIGN/LIBRARY_DESCRIPTOR/LIBRARY_LAYOUT/PAIRED",
            "ORIENTATION",
        ),
        "library_layout_length": (
            "./DESIGN/LIBRARY_DESCRIPTOR/LIBRARY_LAYOUT/PAIRED",
            "NOMINAL_LENGTH",
        ),
        "library_layout_sdev": (
            "./DESIGN/LIBRARY_DESCRIPTOR/LIBRARY_LAYOUT/PAIRED",
            "NOMINAL_SDEV",
        ),
        "pooling_stategy": ("./DESIGN/LIBRARY_DESCRIPTOR/POOLING_STRATEGY", "text"),
        "library_construction_protocol": (
            "./DESIGN/LIBRARY_DESCRIPTOR/LIBRARY_CONSTRUCTION_PROTOCOL",
            "text",
        ),
        "platform": ("./PLATFORM", "child", "tag"),
        "sample_accession": (".//SAMPLE_DESCRIPTOR", "accession"),
        "instrument_model": ("./PLATFORM/*/INSTRUMENT_MODEL", "text"),
    }

    d.update(_process_path_map(xml, path_map))
    d.update(_parse_identifiers(xml.find("IDENTIFIERS"), "experiment"))

    d.update(_parse_attributes(xml.find("EXPERIMENT_ATTRIBUTES")))
    d.update(_parse_links(xml.find("EXPERIMENT_LINKS")))
    d = try_update(d, _parse_run_reads(xml.find(".//SPOT_DESCRIPTOR")))
    return d


def parse_sample(xml):
    """Parse an SRA xml SAMPLE element

    Parameters
    ----------
    xml: an xml.etree Element

    Returns
    -------
    A dict object parsed from the XML
    """

    d = {}
    d.update(xml.attrib)
    path_map = {
        "title": (".//TITLE", "text"),
        "organism": (".//SCIENTIFIC_NAME", "text"),
        "description": (".//DESCRIPTION", "text"),
    }
    d.update(_process_path_map(xml, path_map))
    d.update(_parse_identifiers(xml.find("IDENTIFIERS"), "sample"))
    d.update(_parse_attributes(xml.find("SAMPLE_ATTRIBUTES")))
    d.update(_parse_links(xml.find("SAMPLE_LINKS")))

    for elem in xml.iter():
        if elem.tag == "TAXON_ID":
            d["taxon_id"] = int(elem.text)

    return d


def _parse_run_reads(node):
    """Parse reads from runs."""
    d = dict()
    try:
        d["spot_length"] = int(node.find(".//SPOT_LENGTH").text)
    except Exception as e:
        d["spot_length"] = 0
    d["reads"] = []
    if node is None:
        # No read statistics present
        return d
    readrecs = node.findall(".//READ_SPEC")
    d["nreads"] = len(readrecs)

    for read in readrecs:
        r = {}
        try:
            r["read_index"] = int(read.find("./READ_INDEX").text)
        except:
            pass
        try:
            r["read_class"] = read.find("./READ_CLASS").text
        except:
            pass
        try:
            r["read_type"] = read.find("./READ_TYPE").text
        except:
            pass
        try:
            r["base_coord"] = int(read.find("./BASE_COORD").text)
        except:
            pass
        d["reads"].append(r)

    return d


def _parse_run_qualities(node):
    """Parse the quality stats, if available, from RUN"""
    d = dict()
    d["qualities"] = []
    qualrecs = node.findall(".//Quality")
    for qual in qualrecs:
        try:
            d["qualities"].append(
                {"quality": int(qual.get("value")), "count": int(qual.get("count"))}
            )
        except:
            pass

    return d


def _parse_taxon(node):
    """Parse taxonomy informaiton."""

    def crawl(node, d=[]):
        for i in node:
            rank = i.get("rank", "Unkown")
            parent = None
            if node.get("tax_id") is not None:
                parent = int(node.get("tax_id"))
            d.append(
                {
                    "rank": rank,
                    "name": i.get("name").replace(".", "_").replace("$", ""),
                    "parent": parent,
                    "total_count": int(i.get("total_count")),
                    "self_count": int(i.get("self_count")),
                    "tax_id": int(i.get("tax_id")),
                }
            )

            if len(list(i)) > 0:
                d = d + crawl(i)
        return d

    try:
        d = {
            "tax_analysis": {
                "nspot_analyze": node.get("analyzed_spot_count"),
                "total_spots": node.get("total_spot_count"),
                "mapped_spots": node.get("identified_spot_count"),
                "tax_counts": crawl(node),
            }
        }
    except AttributeError:
        # No tax_analysis node
        return {}

    try:
        if d["tax_analysis"]["nspot_analyze"] is not None:
            d["tax_analysis"]["nspot_analyze"] = int(d["tax_analysis"]["nspot_analyze"])
    except:
        logger.debug("Non integer count: nspot_analyze")
        logger.debug(d["tax_analysis"]["nspot_analyze"])
        d["tax_analysis"]["nspot_analyze"] = None

    try:
        if d["tax_analysis"]["total_spots"] is not None:
            d["tax_analysis"]["total_spots"] = int(d["tax_analysis"]["total_spots"])
    except:
        logger.debug("Non integer count: total_spots")
        logger.debug(d["tax_analysis"]["total_spots"])
        d["tax_analysis"]["total_spots"] = None

    try:
        if d["tax_analysis"]["mapped_spots"] is not None:
            d["tax_analysis"]["mapped_spots"] = int(d["tax_analysis"]["mapped_spots"])
    except:
        logger.debug("Non integer count: mapped_spots")
        logger.debug(d["tax_analysis"]["mapped_spots"])
        d["tax_analysis"]["mapped_spots"] = None

    return d


def _safe_add_text_element(d, key, elem):
    """Add text from an xml element to a dict

    Because not all elements have text elements despite
    their existence in the xml tree, this little
    function checks for text existence and then
    adds the text conditionally. If no text is present,
    the key is not created.

    Parameters
    ----------
    d : dict
        Add the text element to this dict
    key : str
        The key to which to add the text element
    elem : lxml.etree.Element
        From where to extract the text
    """
    txt = elem.text
    if txt is not None:
        d[key] = txt.strip()


def _parse_attributes(xml):
    if xml is None:
        return {}
    """Add attributes to a record

    Parameters
    ----------
    xml: xml.etree.ElementTree.ElementTree.Element
        An xml element of level "EXPERIMENT|STUDY|RUN|SAMPLE_ATTRIBUTES"
    """
    d = defaultdict(list)
    # Iterate over "XXX_ATTRIBUTES"
    for elem in xml:
        try:
            tag = elem.find("./TAG")
            value = elem.find("./VALUE")
            d["attributes"].append({"tag": tag.text, "value": value.text})
        except AttributeError:
            # tag or value missing text, so skip
            pass
    if len(d) == 0:
        d = {}
    return d


def _parse_links(xml):
    """Add attributes to a record

    Parameters
    ----------
    xml: xml.etree.ElementTree.ElementTree.Element
        An xml element of level "EXPERIMENT|STUDY|RUN|SAMPLE_LINKS"
    """
    if xml is None:
        return {}
    d = defaultdict(list)
    # Iterate over "XXX_ATTRIBUTES"
    for elem in xml.findall(".//XREF_LINK"):
        try:
            tag = elem.find("./DB")
            value = elem.find("./ID")
            d["xrefs"].append({"db": tag.text, "id": value.text})
        except AttributeError:
            # tag or value missing text, so skip
            pass
    if len(d) == 0:
        d = {}
    return d


def _get_special_ids(id_rec):
    namespace_map = {
        "geo": "GEO",
        "gds": "GEO_Dataset",
        "pubmed": "pubmed",
        "biosample": "BioSample",
        "bioproject": "BioProject",
    }
    # code below from sramongo/sra.py by jtfear
    # https://github.com/jfear/sramongo/blob/master/sramongo/sra.py
    #
    # Make sure fully formed xref
    try:
        _id = id_rec["id"]
        _db = id_rec["namespace"]
    except Exception:
        return False

    if (_id is None) | (_db is None):
        return False

    # normalize db name
    try:
        norm = _db.strip(" ()[].:").lower()
    except Exception:
        norm = ""

    if norm in namespace_map.keys():
        # Normalize the ids a little
        id_norm = (
            re.sub("geo|gds|bioproject|biosample|pubmed|pmid", "", _id.lower())
            .strip(" :().")
            .upper()
        )
        return namespace_map[norm], id_norm
    else:
        return False


def _parse_identifiers(xml, section):
    """Parse IDENTIFIERS section"""

    d = defaultdict(list)

    for _id in xml:
        if _id.tag == "PRIMARY_ID":
            d[section + "_accession"] = _id.text
        elif _id.tag == "SUBMITTER_ID":
            id_rec = {"namespace": _id.get("namespace"), "id": _id.text}
            d["identifiers"].append(id_rec)
        elif _id.tag == "UUID":
            d["identifiers"].append({"uuid": _id.text})
        else:  # all other id types (secondary, external)
            id_rec = {"namespace": _id.get("namespace"), "id": _id.text}
            d["identifiers"].append(id_rec)
            special = _get_special_ids(id_rec)
            if special:
                d[special[0]] = special[1]
            else:
                d["identifiers"].append(id_rec)
    return d


def _process_path_map(xml, path_map):
    d = {}
    for k, v in path_map.items():
        try:
            # use "text" as second tuple value to
            # get the text value
            if v[1] == "text":
                d[k] = xml.find(v[0]).text
                # use the name of the attribute to
                # get a specific attribute
            elif v[1] == "child":
                child = list(xml.find(v[0]))

                if len(child) > 1:
                    raise Exception("There are too many elements")
                elif v[2] == "text":
                    d[k] = child[0].text
                elif v[2] == "tag":
                    d[k] = child[0].tag
            else:
                d[k] = xml.find(v[0]).get(v[1])
        except Exception as e:
            pass
    return d


def _parse_study_type(xml):
    d = {}
    if xml is None:
        return d
    if xml.get("existing_study_type"):
        d["study_type"] = xml.get("existing_study_type")
    if xml.get("new_study_type"):
        d["study_type"] = xml.get("new_study_type")
    return d


def try_update(d, value):
    try:
        d.update(value)
        return d
    except Exception as e:
        return d


class SRAXMLRecord(object):
    def __init__(self, xml):
        # if(type(xml) is not xml.etree.ElementTree.Element):
        #    raise(TypeError('xml should be of type xml.etree.ElementTree.Element'))
        self.xml = xml
        self.parse_xml()

    def parse_xml(self):
        d = {}
        d = d.update(self.xml.attrib)
        self.data = d


class SRAExperimentRecord(SRAXMLRecord):
    def __init__(self, xml):
        super().__init__(xml)

    def parse_xml(self):
        """Parse an SRA xml EXPERIMENT element"""
        xml = self.xml
        self.data = parse_experiment(xml)


class SRAStudyRecord(SRAXMLRecord):
    def __init__(self, xml):
        super().__init__(xml)

    def parse_xml(self):
        self.data = parse_study(self.xml)


class SRASampleRecord(SRAXMLRecord):
    def __init__(self, xml):
        super().__init__(xml)

    def parse_xml(self):
        self.data = parse_sample(self.xml)


class SRASubmissionRecord(SRAXMLRecord):
    def __init__(self, xml):
        super().__init__(xml)

    def parse_xml(self):
        self.data = parse_submission(self.xml)


class SRARunRecord(SRAXMLRecord):
    def __init__(self, xml):
        super().__init__(xml)

    def parse_xml(self):
        self.data = parse_run(self.xml)


def sra_object_generator(fh):
    """Iterate over objects in an SRA meta_XXX_set xml file

    :param: fh an open filehandle

    :returns: An iterator over SRA objects. Access actual data
        as a dict using Object.data

    """
    from xml.etree import ElementTree as et

    parsers = {
        "study": SRAStudyRecord,
        "sample": SRASampleRecord,
        "run": SRARunRecord,
        "experiment": SRAExperimentRecord,
    }
    validClasses = ["experiment", "run", "study", "sample"]
    for event, element in et.iterparse(fh):
        if (element.tag.lower() in validClasses) and (event == "end"):
            yield parsers.get(element.tag.title().lower())(element)
            element.clear()



"""Simple utility functions for EBI ENA"""
import urllib.request
import xml.etree.ElementTree as etree
import csv
import io
import omicidx.sra.parser


def get_xml_for_accession(accession: str):
    """Get an XML record for a single accession as an ElementTree Node from EBI

    Parameters
    ----------
    accession: str
        Any SRA/ENA accession

    Returns
    -------
    an etree ElementNode parsed from the retrieved XML."""
    url = "https://www.ebi.ac.uk/ena/data/view/{accession}&display=xml&download=xml&filename={accession}.xml"
    url = url.format(accession=accession)
    with urllib.request.urlopen(url) as response:
        return etree.parse(response).getroot()[0]


def get_file_report_for_accession(accession: str):
    """Get a file report from EBI

    Parameters
    ----------
    accession: str
        Any SRA/ENA accession

    Returns
    -------
    a list of rows representing available files for the accession
    """
    url = "https://www.ebi.ac.uk/ena/data/warehouse/filereport?accession={accession}&result=read_run"
    url = url.format(accession=accession)
    print(url)
    response = urllib.request.urlopen(url)
    reader = csv.DictReader(io.StringIO(response.read().decode()), delimiter="\t")
    return list(reader)


def get_parsed_xml_for_accession(accession: str):
    """A parsed SRA record from EBI for a given accession

    Args:
        accession (str): An SRA accession (SRR, SRP, SRS, SRX)

    Returns:
        SRARecord: A parsed SRA accession
    """
    xml = get_xml_for_accession(accession)
    xmlfunc = getattr(omicidx.sra.parser, "{}_xml_iter_parser".format(xml.tag.lower()))
    return xmlfunc(xml)



import gzip
import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from typing import Tuple


def open_file(fname: str, mode: str = "rt"):
    """open a file, including dealing with gzipped files

    Parameters
    ----------
    fname : str
        a filename. If ending in .gz, will use gzip.open().
        Otherwise, a regular open() will be used.

    mode : str
        The file opening mode.
    """
    if fname.endswith(".gz"):
        f = gzip.open(fname, mode)
    else:
        f = open(fname, mode)
    return f


def requests_retry_session(
    retries: int = 3,
    backoff_factor: float = 0.3,
    status_forcelist: Tuple[int] = (500, 502, 504),
    session: requests.Session = None,
) -> requests.Session:
    """Get a requests session that handles retries

        Parameters
        ==========
        retries: int
            Number of retries allowed
        backoff_factor: float
            Implements geometric backoff
        status_forcelist: Tuple[int]
            Status codes to retry for
        session: requests.Session
            Supply a session that will be modified, or
            a new session will be created.

        Returns
        =======
        A new requests.Session with retries enforced.
        ```
    t0 = time.time()
    try:
        response = requests_retry_session().get(
            'http://httpbin.org/delay/10',
            timeout=5
        )
    except Exception as x:
        print('It failed :(', x.__class__.__name__)
    else:
        print('It eventually worked', response.status_code)
    finally:
        t1 = time.time()
        print('Took', t1 - t0, 'seconds')
        ```
    """

    session = session or requests.Session()
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session



import argparse
from .sra.parser import sra_object_generator
import urllib.request
import orjson
import gzip
import tempfile
from upath import UPath
import shutil

def cli():
    parser = argparse.ArgumentParser()
    parser.add_argument("url", help="URL to parse")
    parser.add_argument("outfile", help="Output file")

    args = parser.parse_args()

    # while it is technically possible to stream from the URL
    # the safer way is to download the file first
    with tempfile.NamedTemporaryFile() as tmpfile:

        # download the file to a temporary file
        shutil.copyfileobj(UPath(args.url).open("rb"), tmpfile)
        # read the file and write to the output file
        tmpfile.seek(0)
        with gzip.open(tmpfile, 'rb') as fh:
            with UPath(args.outfile).open('wb', compression='gzip') as outfile:

                # iterate over the objects and write them to the output file
                for obj in sra_object_generator(fh):
                    outfile.write(orjson.dumps(obj.data) + b"\n")

if __name__ == "__main__":
    cli()